/******************************************************************************
    Copyright (c) 1996-2000 Synopsys, Inc.    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/tapinprogram.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include "syntax.h"
#include <stdlib.h>
#include <math.h>
#include "mymalloc.h"
#include "LibEnumMap.h"
#include "LibGroup.h"
#include "LibParse.h"
#include <float.h>

static int errcount;

static void check_group_correspondence(LibGroup* g, LibGroup* parent,
				       libsynt_group_info * gi);
static void check_attr_correspondence(LibAttr* attr,
				      libsynt_attribute_info * ai);

LibMsgHandlerT MsgPrinter;

static char EB[8000];		/* just so I don't have to declare buffers all over the place,
				   to prepare output for output error messages */

static int
isa_formula(const char *str)
{
  /* a crude approximation at a formula-- the real best solution here would be to
     have a general parser for arithmetic expressions; constant arithmetic expressions? */
  const char *p = str;

  while (*p) {
    if ((*p >= '0' && *p <= '9') || *p == ' ' || *p == '\t' || *p == '*'
	|| *p == '/' || *p == '+' || *p == '-' || *p == 'e' || *p == 'E'
	|| *p == '.') {
      p++;
    } else if (*p == 'V' && *(p + 1) == 'D' && *(p + 2) == 'D') {
      p += 3;
    } else {
      break;
    }
  }
  if (*p) {
    return 0;
  } else {
    return 1;
  }
}

typedef struct var_list
{
  char *name;
  struct var_list *next;

} var_list;

static var_list *master_var_list = 0;

static void
add_varlist(char *name)
{
  int found = 0;
  var_list *p;

  for (p = master_var_list; p; p = p->next) {
    if (!strcmp(name, p->name)) {
      found = 1;
      break;
    }
  }
  if (!found) {
    p = (var_list *) my_calloc(sizeof(var_list), 1);
    p->name = (char *)my_malloc(strlen(name) + 1);
    strcpy(p->name, name);
    p->next = master_var_list;
    master_var_list = p;
  }
}

static void
destroy_varlist(void)
{
  var_list *p, *p1;

  for (p = master_var_list; p; p = p1) {
    p1 = p->next;
    my_free(p->name);
    my_free(p);
  }
  master_var_list = 0;
}

#if 0
static void
print_var_list(void)
{
  var_list *p;
  LibErrorT err;

  (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, "---- var list:");
  for (p = master_var_list; p; p = p->next) {
    (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, p->name);
  }
  (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, "----");
}
#endif


LibErrorT
syntax_check(LibGroup* lib)
{

  extern libsynt_technology libsynt_techs[];
  extern int libsynt_tech_num;
  LibAttr* attr;
  LibErrorT err;
  LibValType vtype;
  LibValIter viter;
  LibVal* v;
  int found, first, i = 0, dmln, tcln;
  const char* model;
  const char* tech;
  const char* tcfn=NULL;
  const char* dmfn=NULL;
  LibAttrType atype;

  errcount = 0;

  /* one thing upon which all others hinge!

     First: this top group must be a library.

     We'll quick scan for tech and model...

   */
  MsgPrinter = LibParse::msgHandler();

  if (lib->type() != "library") {
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR,
		   "Unrecognized library file - the top level group should be of type 'library' ");

    return LIB_SYNTAX_ERROR;
  }

  attr = lib->findAttr(K_technology, err);
  tech = NULL;
  if (attr) {
    atype = attr->type();
    tcln = attr->lineno();
    tcfn = attr->filename();

    if (atype != LIBERTY__COMPLEX) {

      sprintf(EB,
	      "The technology attribute specified at %s:%d is a simple attribute and should be a complex attribute!",
              tcfn, tcln);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      vtype = attr->value()->type();
      switch (vtype) {
      case LIBERTY__VAL_STRING:
	tech = attr->value()->string_val();
	break;

      case LIBERTY__VAL_DOUBLE:
      case LIBERTY__VAL_INT:
      case LIBERTY__VAL_BOOLEAN:
	errcount++;
	sprintf(EB,
		"Not only is the technology attribute type wrong, so is the value type! "
		"It should be a string, but there's a value of some other type here!");
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	break;

      default:
	break;
      }
    } else {
      found = 0;
      first = 1;

      viter = attr->valueIter();
      while ((v = viter++)) {
	switch (v->type()) {
	case LIBERTY__VAL_STRING:
	  tech = v->string_val();
	  if (first)
	    first = 0;
	  else {
	    sprintf(EB,
		    "Found a string argument, but it was not the first and only argument of technology!");
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    errcount++;
	  }
	  found = 1;
	  break;

	case LIBERTY__VAL_DOUBLE:
	case LIBERTY__VAL_INT:
	case LIBERTY__VAL_BOOLEAN:
	  sprintf(EB,
		  "The technology attribute at %s:%d should only have one string value; More than one, or non-string attributes detected!",
		  tcfn, tcln);
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  errcount++;
	  break;

        default:
	  break;
	}
      }
      if (!found) {
	sprintf(EB,
		"The technology attribute at %s:%d should only have one string value; No string value was  detected!",
		tcfn, tcln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      } else {
	sprintf(EB, "The technology %s was specified.", tech);
	(*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, EB);
      }
    }
  } else {
    sprintf(EB,
	    "No technology attribute specified in the library; the technology 'cmos' is assumed.");
    (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, EB);
    tech = "cmos";
    tcln = -1;
  }

  attr = lib->findAttr("delay_model", err);
  model = "generic_cmos";  // generic_cmos assumed if delay_model not correctly specified
  if (attr) {
    dmln = attr->lineno();
    dmfn = attr->filename();
    atype = attr->type();
    if (atype != LIBERTY__SIMPLE) {
      sprintf(EB,
	      "The delay_model attribute was not a simple attribute. Ignoring the specification. 'generic_cmos' will be assumed.");
      (*MsgPrinter) (LIB_SEVERITY_WARN, LIB_NO_ERROR, EB);
      errcount++;
    } else {
      vtype = attr->value()->type();
      switch (vtype) {
      case LIBERTY__VAL_STRING:
	model = attr->value()->string_val();
	sprintf(EB, "delay_model specified was %s.", model);
	(*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, EB);
	break;

      case LIBERTY__VAL_DOUBLE:
      case LIBERTY__VAL_INT:
      case LIBERTY__VAL_BOOLEAN:
	sprintf(EB,
		"The delay_model attribute at %s:%d was not a simple attribute of string type. 'generic_cmos' will be assumed.",
		dmfn, dmln);
	(*MsgPrinter) (LIB_SEVERITY_WARN, LIB_NO_ERROR, EB);
	errcount++;
	break;

      default:
        break;
      }
    }
  } else {
    (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR,
		   "No delay_model attribute specified in the library; the delay_model 'generic_cmos' is assumed.");
    dmln = -1;
  }
  /* right at this moment, I need to come up with a better convention for naming the desc files.... */
  /* find the cmos/cmos model for now */
  found = 0;

  if (dmln > 0 && tcln > 0 && dmln < tcln) {
    sprintf(EB,
	    "The delay_model should be declared after the technology, but is not. Check lines %s:%d and %s:%d.",
	    tcfn, tcln, dmfn, dmln);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }

  if (!strcmp(tech, "cmos"))
    for (i = 0; i < libsynt_tech_num; i++) {
      if (!strcmp(libsynt_techs[i].name, model)) {
	found = 1;
	break;
      }
    }
  if (!found)
    for (i = 0; i < libsynt_tech_num; i++) {
      if (!strcmp(libsynt_techs[i].name, tech)) {
	found = 1;
	break;
      }
    }
  if (!found) {
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR,
		   "I cannot check the syntax of this particular technology/delay_model combination!");

    (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR,
		   " Here is a list of technologies/delay_models I can handle:");
    for (i = 0; i < libsynt_tech_num; i++) {
      sprintf(EB, "  %s", libsynt_techs[i].name);
      (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, EB);
    }
    return LIB_SYNTAX_ERROR;
  }

  sprintf(EB, "Using the %s syntax tables...", libsynt_techs[i].name);
  (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_NO_ERROR, EB);

  check_group_correspondence(lib, NULL, libsynt_techs[i].lib);

  if (errcount > 0)
    return LIB_SYNTAX_ERROR;
  else
    return LIB_NO_ERROR;
}

static void
enter_define_cell_area(LibAttr* a)
{
  LibVal *v;
  LibValIter viter;
  const char* string;

  /* the object here is to add a define to the database;
     but in such a way so it will not show up if we print the database! */

  /* first, get the first argument */

  /* should be a string with a list of comma/space sep'd numbers */
  if (a->type() != LIBERTY__COMPLEX) {
    return;			/* this will already be reported */
  }

  viter = a->valueIter();
  v = viter++;

  if (!v) {
    sprintf(EB, "%s:%d: The 'define_cell_area' attribute has no values.",
	    a->filename(), a->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    return;
  }

  if (v->type() != LIBERTY__VAL_STRING) {
    sprintf(EB, "%s:%d: The 'define_cell_area' attribute should be a string.",
	    a->filename(), a->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    return;
  }

  string = v->string_val();
  if (!string || !string[0]) {
    sprintf(EB, "%s:%d: The 'define_cell_area' attribute is an empty string.",
	    a->filename(), a->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    return;
  }
  LibParse::group()->createFloatingDefineForCellArea(string);
}

static void
attribute_specific_checks(LibAttr* a)
{
  attr_enum attrtype = (attr_enum)a->typeID();

  switch (attrtype) {
  case LIBERTY_ATTRENUM_input_pins:
    break;

  case LIBERTY_ATTRENUM_fpga_timing_type:
    break;

  case LIBERTY_ATTRENUM_fpga_degenerate_output:
    break;

  case LIBERTY_ATTRENUM_xnf_schnm:
    break;

  case LIBERTY_ATTRENUM_xnf_device_name:
    break;

  case LIBERTY_ATTRENUM_fpga_min_degen_input_size:
    break;

  case LIBERTY_ATTRENUM_fpga_max_degen_input_size:
    break;

  case LIBERTY_ATTRENUM_fpga_lut_output:
    break;

  case LIBERTY_ATTRENUM_fpga_lut_insert_before_sequential:
    break;

  case LIBERTY_ATTRENUM_fpga_family:
    break;

  case LIBERTY_ATTRENUM_fpga_complex_degenerate:
    break;

  case LIBERTY_ATTRENUM_fpga_cell_type:
    break;

  case LIBERTY_ATTRENUM_fpga_bridge_inputs:
    break;

  case LIBERTY_ATTRENUM_fpga_prefer_undegenerated_gates:
    break;

  case LIBERTY_ATTRENUM_fpga_allow_duplicate_degenerated_gates:
    break;

  case LIBERTY_ATTRENUM_force_inverter_removal:
    break;

  case LIBERTY_ATTRENUM_fanout_resistance:
    break;

  case LIBERTY_ATTRENUM_fanout_capacitance:
    break;

  case LIBERTY_ATTRENUM_fanout_area:
    break;

  case LIBERTY_ATTRENUM_wire_load_from_area:
    break;

  case LIBERTY_ATTRENUM_fanout_length:
    break;

  case LIBERTY_ATTRENUM_slope:
    break;

  case LIBERTY_ATTRENUM_resistance:
    break;

  case LIBERTY_ATTRENUM_slower_factor:
    break;

  case LIBERTY_ATTRENUM_faster_factor:
    break;

  case LIBERTY_ATTRENUM_default_power_rail:
    break;

  case LIBERTY_ATTRENUM_vomin:
    break;

  case LIBERTY_ATTRENUM_vomax:
    break;

  case LIBERTY_ATTRENUM_vol:
    break;

  case LIBERTY_ATTRENUM_voh:
    break;

  case LIBERTY_ATTRENUM_power_rail:
    break;

  case LIBERTY_ATTRENUM_voltage:
    break;

  case LIBERTY_ATTRENUM_tree_type:
    break;

  case LIBERTY_ATTRENUM_temperature:
    break;

  case LIBERTY_ATTRENUM_process:
    break;

  case LIBERTY_ATTRENUM_short:
    break;

  case LIBERTY_ATTRENUM_cell_name:
    break;

  case LIBERTY_ATTRENUM_variable_3:
    break;

  case LIBERTY_ATTRENUM_vimin:
    break;

  case LIBERTY_ATTRENUM_vimax:
    break;

  case LIBERTY_ATTRENUM_vil:
    break;

  case LIBERTY_ATTRENUM_vih:
    break;

  case LIBERTY_ATTRENUM_variable_2:
    break;

  case LIBERTY_ATTRENUM_variable_1:
    break;

  case LIBERTY_ATTRENUM_downto:
    break;

  case LIBERTY_ATTRENUM_data_type:
    break;

  case LIBERTY_ATTRENUM_bit_width:
    break;

  case LIBERTY_ATTRENUM_bit_to:
    break;

  case LIBERTY_ATTRENUM_bit_from:
    break;

  case LIBERTY_ATTRENUM_base_type:
    break;

  case LIBERTY_ATTRENUM_table:
    break;

  case LIBERTY_ATTRENUM_force_11:
    break;

  case LIBERTY_ATTRENUM_force_10:
    break;

  case LIBERTY_ATTRENUM_force_01:
    break;

  case LIBERTY_ATTRENUM_force_00:
    break;

  case LIBERTY_ATTRENUM_tracks:
    break;

  case LIBERTY_ATTRENUM_total_track_area:
    break;

  case LIBERTY_ATTRENUM_word_width:
    break;

  case LIBERTY_ATTRENUM_type:
    break;

  case LIBERTY_ATTRENUM_row_address:
    break;

  case LIBERTY_ATTRENUM_column_address:
    break;

  case LIBERTY_ATTRENUM_address_width:
    break;

  case LIBERTY_ATTRENUM_value:
    break;

  case LIBERTY_ATTRENUM_enable_also:
    break;

  case LIBERTY_ATTRENUM_data_in:
    break;

  case LIBERTY_ATTRENUM_shifts:
    break;

  case LIBERTY_ATTRENUM_edges:
    break;

  case LIBERTY_ATTRENUM_multiplied_by:
    break;

  case LIBERTY_ATTRENUM_master_pin:
    break;

  case LIBERTY_ATTRENUM_invert:
    break;

  case LIBERTY_ATTRENUM_duty_cycle:
    break;

  case LIBERTY_ATTRENUM_divided_by:
    break;

  case LIBERTY_ATTRENUM_clock_pin:
    break;

  case LIBERTY_ATTRENUM_preset:
    break;

  case LIBERTY_ATTRENUM_next_state:
    break;

  case LIBERTY_ATTRENUM_clocked_on_also:
    break;

  case LIBERTY_ATTRENUM_clear_preset_var2:
    break;

  case LIBERTY_ATTRENUM_clear_preset_var1:
    break;

  case LIBERTY_ATTRENUM_clear:
    break;

  case LIBERTY_ATTRENUM_bus_type:
    break;

  case LIBERTY_ATTRENUM_tdisable:
    break;

  case LIBERTY_ATTRENUM_edge_type:
    break;

  case LIBERTY_ATTRENUM_rise_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_rise_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_mode:
    break;

  case LIBERTY_ATTRENUM_fall_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_fall_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_when_start:
    break;

  case LIBERTY_ATTRENUM_when_end:
    break;

  case LIBERTY_ATTRENUM_timing_type:
    break;

  case LIBERTY_ATTRENUM_timing_sense:
    break;

  case LIBERTY_ATTRENUM_steady_state_resistance_low_min:
    break;

  case LIBERTY_ATTRENUM_steady_state_resistance_low_max:
    break;

  case LIBERTY_ATTRENUM_steady_state_resistance_high_min:
    break;

  case LIBERTY_ATTRENUM_steady_state_resistance_high_max:
    break;

  case LIBERTY_ATTRENUM_steady_state_resistance_float_min:
    break;

  case LIBERTY_ATTRENUM_steady_state_resistance_float_max:
    break;

  case LIBERTY_ATTRENUM_slope_rise:
    break;

  case LIBERTY_ATTRENUM_slope_fall:
    break;

  case LIBERTY_ATTRENUM_setup_coefficient:
    break;

  case LIBERTY_ATTRENUM_sdf_edges:
    break;

  case LIBERTY_ATTRENUM_sdf_cond_start:
    break;

  case LIBERTY_ATTRENUM_sdf_cond_end:
    break;

  case LIBERTY_ATTRENUM_rise_resistance:
    break;

  case LIBERTY_ATTRENUM_related_output_pin:
    break;

  case LIBERTY_ATTRENUM_related_bus_equivalent:
    break;

  case LIBERTY_ATTRENUM_intrinsic_rise:
    break;

  case LIBERTY_ATTRENUM_intrinsic_fall:
    break;

  case LIBERTY_ATTRENUM_hold_coefficient:
    break;

  case LIBERTY_ATTRENUM_fall_resistance:
    break;

  case LIBERTY_ATTRENUM_edge_rate_sensitivity_r1:
    break;

  case LIBERTY_ATTRENUM_edge_rate_sensitivity_r0:
    break;

  case LIBERTY_ATTRENUM_edge_rate_sensitivity_f1:
    break;

  case LIBERTY_ATTRENUM_edge_rate_sensitivity_f0:
    break;

  case LIBERTY_ATTRENUM_default_timing:
    break;

  case LIBERTY_ATTRENUM_lsi_pad:
    break;

  case LIBERTY_ATTRENUM_fsim_map:
    break;

  case LIBERTY_ATTRENUM_constraint:
    break;

  case LIBERTY_ATTRENUM_sdf_cond:
    break;

  case LIBERTY_ATTRENUM_constraint_low:
    break;

  case LIBERTY_ATTRENUM_constraint_high:
    break;

  case LIBERTY_ATTRENUM_enable:
    break;

  case LIBERTY_ATTRENUM_clocked_on:
    break;

  case LIBERTY_ATTRENUM_address:
    break;

  case LIBERTY_ATTRENUM_when:
    break;

  case LIBERTY_ATTRENUM_switching_together_group:
    break;

  case LIBERTY_ATTRENUM_switching_interval:
    break;

  case LIBERTY_ATTRENUM_rising_together_group:
    break;

  case LIBERTY_ATTRENUM_related_rising_pin:
    break;

  case LIBERTY_ATTRENUM_related_outputs:
    break;

  case LIBERTY_ATTRENUM_related_inputs:
    break;

  case LIBERTY_ATTRENUM_related_input:
    break;

  case LIBERTY_ATTRENUM_related_falling_pin:
    break;

  case LIBERTY_ATTRENUM_falling_together_group:
    break;

  case LIBERTY_ATTRENUM_equal_or_opposite_output:
    break;

  case LIBERTY_ATTRENUM_values:
    break;

  case LIBERTY_ATTRENUM_intermediate_values:
    break;

  case LIBERTY_ATTRENUM_index_3:
    break;

  case LIBERTY_ATTRENUM_index_2:
    break;

  case LIBERTY_ATTRENUM_index_1:
    break;

  case LIBERTY_ATTRENUM_related_pin:
    break;

  case LIBERTY_ATTRENUM_related_bus_pins:
    break;

  case LIBERTY_ATTRENUM_members:
    break;

  case LIBERTY_ATTRENUM_input_map_shift:
    break;

  case LIBERTY_ATTRENUM_x_function:
    break;

  case LIBERTY_ATTRENUM_three_state:
    break;

  case LIBERTY_ATTRENUM_test_output_only:
    break;

  case LIBERTY_ATTRENUM_state_function:
    break;

  case LIBERTY_ATTRENUM_slew_control:
    break;

  case LIBERTY_ATTRENUM_signal_type:
    break;

  case LIBERTY_ATTRENUM_rise_time_before_threshold:
    break;

  case LIBERTY_ATTRENUM_rise_time_after_threshold:
    break;

  case LIBERTY_ATTRENUM_rise_current_slope_before_threshold:
    break;

  case LIBERTY_ATTRENUM_rise_current_slope_after_threshold:
    break;

  case LIBERTY_ATTRENUM_rise_capacitance:
    break;

  case LIBERTY_ATTRENUM_reference_capacitance:
    break;

  case LIBERTY_ATTRENUM_pulling_resistance:
    break;

  case LIBERTY_ATTRENUM_pulling_current:
    break;

  case LIBERTY_ATTRENUM_primary_output:
    break;

  case LIBERTY_ATTRENUM_prefer_tied:
    break;

  case LIBERTY_ATTRENUM_pin_func_type:
    break;

  case LIBERTY_ATTRENUM_output_voltage:
    break;

  case LIBERTY_ATTRENUM_output_signal_level:
    break;

  case LIBERTY_ATTRENUM_nextstate_type:
    break;

  case LIBERTY_ATTRENUM_multicell_pad_pin:
    break;

  case LIBERTY_ATTRENUM_min_transition:
    break;

  case LIBERTY_ATTRENUM_min_pulse_width_low:
    break;

  case LIBERTY_ATTRENUM_min_pulse_width_high:
    break;

  case LIBERTY_ATTRENUM_min_period:
    break;

  case LIBERTY_ATTRENUM_min_fanout:
    break;

  case LIBERTY_ATTRENUM_min_capacitance:
    break;

  case LIBERTY_ATTRENUM_max_transition:
    break;

  case LIBERTY_ATTRENUM_max_time_borrow:
    break;

  case LIBERTY_ATTRENUM_max_fanout:
    break;

  case LIBERTY_ATTRENUM_max_capacitance:
    break;

  case LIBERTY_ATTRENUM_is_pad:
    break;

  case LIBERTY_ATTRENUM_inverted_output:
    break;

  case LIBERTY_ATTRENUM_internal_node:
    break;

  case LIBERTY_ATTRENUM_input_voltage:
    break;

  case LIBERTY_ATTRENUM_input_signal_level:
    break;

  case LIBERTY_ATTRENUM_input_map:
    break;

  case LIBERTY_ATTRENUM_hysteresis:
    break;

  case LIBERTY_ATTRENUM_function:
    break;

  case LIBERTY_ATTRENUM_fault_model:
    break;

  case LIBERTY_ATTRENUM_fanout_load:
    break;

  case LIBERTY_ATTRENUM_fall_time_before_threshold:
    break;

  case LIBERTY_ATTRENUM_fall_time_after_threshold:
    break;

  case LIBERTY_ATTRENUM_fall_current_slope_before_threshold:
    break;

  case LIBERTY_ATTRENUM_fall_current_slope_after_threshold:
    break;

  case LIBERTY_ATTRENUM_fall_capacitance:
    break;

  case LIBERTY_ATTRENUM_edge_rate_rise:
    break;

  case LIBERTY_ATTRENUM_edge_rate_load_rise:
    break;

  case LIBERTY_ATTRENUM_edge_rate_load_fall:
    break;

  case LIBERTY_ATTRENUM_edge_rate_fall:
    break;

  case LIBERTY_ATTRENUM_edge_rate_breakpoint_r1:
    break;

  case LIBERTY_ATTRENUM_edge_rate_breakpoint_r0:
    break;

  case LIBERTY_ATTRENUM_edge_rate_breakpoint_f1:
    break;

  case LIBERTY_ATTRENUM_edge_rate_breakpoint_f0:
    break;

  case LIBERTY_ATTRENUM_driver_type:
    break;

  case LIBERTY_ATTRENUM_drive_current:
    break;

  case LIBERTY_ATTRENUM_direction:
    break;

  case LIBERTY_ATTRENUM_dcm_timing:
    break;

  case LIBERTY_ATTRENUM_connection_class:
    break;

  case LIBERTY_ATTRENUM_complementary_pin:
    break;

  case LIBERTY_ATTRENUM_clock_gate_test_pin:
    break;

  case LIBERTY_ATTRENUM_clock_gate_out_pin:
    break;

  case LIBERTY_ATTRENUM_clock_gate_obs_pin:
    break;

  case LIBERTY_ATTRENUM_clock_gate_enable_pin:
    break;

  case LIBERTY_ATTRENUM_clock_gate_clock_pin:
    break;

  case LIBERTY_ATTRENUM_clock:
    break;

  case LIBERTY_ATTRENUM_capacitance:
    break;

  case LIBERTY_ATTRENUM_rail_connection:
    break;

  case LIBERTY_ATTRENUM_pin_opposite:
    break;

  case LIBERTY_ATTRENUM_pin_equal:
    break;

  case LIBERTY_ATTRENUM_vhdl_name:
    break;

  case LIBERTY_ATTRENUM_use_for_size_only:
    break;

  case LIBERTY_ATTRENUM_single_bit_degenerate:
    break;

  case LIBERTY_ATTRENUM_set_node:
    break;

  case LIBERTY_ATTRENUM_scan_group:
    break;

  case LIBERTY_ATTRENUM_scaling_factors:
    break;

  case LIBERTY_ATTRENUM_preferred:
    break;

  case LIBERTY_ATTRENUM_pad_type:
    break;

  case LIBERTY_ATTRENUM_pad_cell:
    break;

  case LIBERTY_ATTRENUM_observe_node:
    break;

  case LIBERTY_ATTRENUM_mpm_name:
    break;

  case LIBERTY_ATTRENUM_mpm_libname:
    break;

  case LIBERTY_ATTRENUM_map_only:
    break;

  case LIBERTY_ATTRENUM_is_clock_gating_cell:
    break;

  case LIBERTY_ATTRENUM_interface_timing:
    break;

  case LIBERTY_ATTRENUM_handle_negative_constraint:
    break;

  case LIBERTY_ATTRENUM_geometry_print:
    break;

  case LIBERTY_ATTRENUM_dont_use:
    break;

  case LIBERTY_ATTRENUM_dont_touch:
    break;

  case LIBERTY_ATTRENUM_dont_fault:
    break;

  case LIBERTY_ATTRENUM_contention_condition:
    break;

  case LIBERTY_ATTRENUM_clock_gating_integrated_cell:
    break;

  case LIBERTY_ATTRENUM_cell_leakage_power:
    break;

  case LIBERTY_ATTRENUM_cell_footprint:
    break;

  case LIBERTY_ATTRENUM_auxiliary_pad_cell:
    break;

  case LIBERTY_ATTRENUM_area:
    break;

  case LIBERTY_ATTRENUM_technology:
    break;

  case LIBERTY_ATTRENUM_routing_layers:
    break;

  case LIBERTY_ATTRENUM_piece_define:
    break;

  case LIBERTY_ATTRENUM_library_features:
    break;

  case LIBERTY_ATTRENUM_define_cell_area:
    enter_define_cell_area(a);
    break;

  case LIBERTY_ATTRENUM_capacitive_load_unit:
    break;

  case LIBERTY_ATTRENUM_voltage_unit:
    break;

  case LIBERTY_ATTRENUM_timing_report:
    break;

  case LIBERTY_ATTRENUM_time_unit:
    break;

  case LIBERTY_ATTRENUM_slew_upper_threshold_pct_rise:
    break;

  case LIBERTY_ATTRENUM_slew_upper_threshold_pct_fall:
    break;

  case LIBERTY_ATTRENUM_slew_lower_threshold_pct_rise:
    break;

  case LIBERTY_ATTRENUM_slew_lower_threshold_pct_fall:
    break;

  case LIBERTY_ATTRENUM_slew_derate_from_library:
    break;

  case LIBERTY_ATTRENUM_simulation:
    break;

  case LIBERTY_ATTRENUM_revision:
    break;

  case LIBERTY_ATTRENUM_pulling_resistance_unit:
    break;

  case LIBERTY_ATTRENUM_preferred_output_pad_voltage:
    break;

  case LIBERTY_ATTRENUM_preferred_output_pad_slew_rate_control:
    break;

  case LIBERTY_ATTRENUM_preferred_input_pad_voltage:
    break;

  case LIBERTY_ATTRENUM_piece_type:
    break;

  case LIBERTY_ATTRENUM_output_threshold_pct_rise:
    break;

  case LIBERTY_ATTRENUM_output_threshold_pct_fall:
    break;

  case LIBERTY_ATTRENUM_nom_voltage:
    break;

  case LIBERTY_ATTRENUM_nom_temperature:
    break;

  case LIBERTY_ATTRENUM_nom_process:
    break;

  case LIBERTY_ATTRENUM_lsi_rounding_digit:
    break;

  case LIBERTY_ATTRENUM_lsi_rounding_cutoff:
    break;

  case LIBERTY_ATTRENUM_lsi_pad_rise:
    break;

  case LIBERTY_ATTRENUM_lsi_pad_fall:
    break;

  case LIBERTY_ATTRENUM_leakage_power_unit:
    break;

  case LIBERTY_ATTRENUM_key_version:
    break;

  case LIBERTY_ATTRENUM_key_seed:
    break;

  case LIBERTY_ATTRENUM_key_file:
    break;

  case LIBERTY_ATTRENUM_key_feature:
    break;

  case LIBERTY_ATTRENUM_key_bit:
    break;

  case LIBERTY_ATTRENUM_k_volt_wire_res:
    break;

  case LIBERTY_ATTRENUM_k_volt_wire_cap:
    break;

  case LIBERTY_ATTRENUM_k_volt_slope_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_slope_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_skew_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_skew_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_setup_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_setup_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_rise_transition:
    break;

  case LIBERTY_ATTRENUM_k_volt_rise_propagation:
    break;

  case LIBERTY_ATTRENUM_k_volt_rise_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_k_volt_rise_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_k_volt_removal_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_removal_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_recovery_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_recovery_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_pin_cap:
    break;

  case LIBERTY_ATTRENUM_k_volt_nochange_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_nochange_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_min_pulse_width_low:
    break;

  case LIBERTY_ATTRENUM_k_volt_min_pulse_width_high:
    break;

  case LIBERTY_ATTRENUM_k_volt_min_period:
    break;

  case LIBERTY_ATTRENUM_k_volt_intrinsic_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_intrinsic_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_internal_power:
    break;

  case LIBERTY_ATTRENUM_k_volt_hold_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_hold_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_fall_transition:
    break;

  case LIBERTY_ATTRENUM_k_volt_fall_propagation:
    break;

  case LIBERTY_ATTRENUM_k_volt_fall_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_k_volt_fall_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_k_volt_drive_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_drive_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_drive_current:
    break;

  case LIBERTY_ATTRENUM_k_volt_cell_rise:
    break;

  case LIBERTY_ATTRENUM_k_volt_cell_leakage_power:
    break;

  case LIBERTY_ATTRENUM_k_volt_cell_fall:
    break;

  case LIBERTY_ATTRENUM_k_volt_cell_degradation:
    break;

  case LIBERTY_ATTRENUM_k_temp_wire_res:
    break;

  case LIBERTY_ATTRENUM_k_temp_wire_cap:
    break;

  case LIBERTY_ATTRENUM_k_temp_slope_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_slope_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_skew_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_skew_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_setup_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_setup_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_rise_transition:
    break;

  case LIBERTY_ATTRENUM_k_temp_rise_propagation:
    break;

  case LIBERTY_ATTRENUM_k_temp_rise_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_k_temp_rise_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_k_temp_removal_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_removal_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_recovery_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_recovery_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_pin_cap:
    break;

  case LIBERTY_ATTRENUM_k_temp_nochange_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_nochange_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_min_pulse_width_low:
    break;

  case LIBERTY_ATTRENUM_k_temp_min_pulse_width_high:
    break;

  case LIBERTY_ATTRENUM_k_temp_min_period:
    break;

  case LIBERTY_ATTRENUM_k_temp_intrinsic_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_intrinsic_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_internal_power:
    break;

  case LIBERTY_ATTRENUM_k_temp_hold_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_hold_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_fall_transition:
    break;

  case LIBERTY_ATTRENUM_k_temp_fall_propagation:
    break;

  case LIBERTY_ATTRENUM_k_temp_fall_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_k_temp_fall_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_k_temp_drive_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_drive_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_drive_current:
    break;

  case LIBERTY_ATTRENUM_k_temp_cell_rise:
    break;

  case LIBERTY_ATTRENUM_k_temp_cell_leakage_power:
    break;

  case LIBERTY_ATTRENUM_k_temp_cell_fall:
    break;

  case LIBERTY_ATTRENUM_k_temp_cell_degradation:
    break;

  case LIBERTY_ATTRENUM_k_process_wire_res:
    break;

  case LIBERTY_ATTRENUM_k_process_wire_cap:
    break;

  case LIBERTY_ATTRENUM_k_process_slope_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_slope_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_skew_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_skew_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_setup_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_setup_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_rise_transition:
    break;

  case LIBERTY_ATTRENUM_k_process_rise_propagation:
    break;

  case LIBERTY_ATTRENUM_k_process_rise_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_k_process_rise_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_k_process_removal_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_removal_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_recovery_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_recovery_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_pin_rise_cap:
    break;

  case LIBERTY_ATTRENUM_k_process_pin_fall_cap:
    break;

  case LIBERTY_ATTRENUM_k_process_pin_cap:
    break;

  case LIBERTY_ATTRENUM_k_process_nochange_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_nochange_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_min_pulse_width_low:
    break;

  case LIBERTY_ATTRENUM_k_process_min_pulse_width_high:
    break;

  case LIBERTY_ATTRENUM_k_process_min_period:
    break;

  case LIBERTY_ATTRENUM_k_process_intrinsic_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_intrinsic_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_internal_power:
    break;

  case LIBERTY_ATTRENUM_k_process_hold_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_hold_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_fall_transition:
    break;

  case LIBERTY_ATTRENUM_k_process_fall_propagation:
    break;

  case LIBERTY_ATTRENUM_k_process_fall_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_k_process_fall_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_k_process_drive_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_drive_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_drive_current:
    break;

  case LIBERTY_ATTRENUM_k_process_cell_rise:
    break;

  case LIBERTY_ATTRENUM_k_process_cell_leakage_power:
    break;

  case LIBERTY_ATTRENUM_k_process_cell_fall:
    break;

  case LIBERTY_ATTRENUM_k_process_cell_degradation:
    break;

  case LIBERTY_ATTRENUM_input_threshold_pct_rise:
    break;

  case LIBERTY_ATTRENUM_input_threshold_pct_fall:
    break;

  case LIBERTY_ATTRENUM_in_place_swap_mode:
    break;

  case LIBERTY_ATTRENUM_em_temp_degradation_factor:
    break;

  case LIBERTY_ATTRENUM_delay_model:
    break;

  case LIBERTY_ATTRENUM_default_wire_load_selection:
    break;

  case LIBERTY_ATTRENUM_default_wire_load_resistance:
    break;

  case LIBERTY_ATTRENUM_default_wire_load_mode:
    break;

  case LIBERTY_ATTRENUM_default_wire_load_capacitance:
    break;

  case LIBERTY_ATTRENUM_default_wire_load_area:
    break;

  case LIBERTY_ATTRENUM_default_wire_load:
    break;

  case LIBERTY_ATTRENUM_default_slope_rise:
    break;

  case LIBERTY_ATTRENUM_default_slope_fall:
    break;

  case LIBERTY_ATTRENUM_default_setup_coefficient:
    break;

  case LIBERTY_ATTRENUM_default_rise_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_default_rise_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_default_reference_capacitance:
    break;

  case LIBERTY_ATTRENUM_default_rc_rise_coefficient:
    break;

  case LIBERTY_ATTRENUM_default_rc_fall_coefficient:
    break;

  case LIBERTY_ATTRENUM_default_output_pin_rise_res:
    break;

  case LIBERTY_ATTRENUM_default_output_pin_fall_res:
    break;

  case LIBERTY_ATTRENUM_default_output_pin_cap:
    break;

  case LIBERTY_ATTRENUM_default_operating_conditions:
    break;

  case LIBERTY_ATTRENUM_default_min_porosity:
    break;

  case LIBERTY_ATTRENUM_default_max_utilization:
    break;

  case LIBERTY_ATTRENUM_default_max_transition:
    break;

  case LIBERTY_ATTRENUM_default_max_fanout:
    break;

  case LIBERTY_ATTRENUM_default_max_capacitance:
    break;

  case LIBERTY_ATTRENUM_default_leakage_power_density:
    break;

  case LIBERTY_ATTRENUM_default_intrinsic_rise:
    break;

  case LIBERTY_ATTRENUM_default_intrinsic_fall:
    break;

  case LIBERTY_ATTRENUM_default_input_pin_cap:
    break;

  case LIBERTY_ATTRENUM_default_inout_pin_rise_res:
    break;

  case LIBERTY_ATTRENUM_default_inout_pin_fall_res:
    break;

  case LIBERTY_ATTRENUM_default_inout_pin_cap:
    break;

  case LIBERTY_ATTRENUM_default_hold_coefficient:
    break;

  case LIBERTY_ATTRENUM_default_fanout_load:
    break;

  case LIBERTY_ATTRENUM_default_fall_pin_resistance:
    break;

  case LIBERTY_ATTRENUM_default_fall_delay_intercept:
    break;

  case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_r1:
    break;

  case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_r0:
    break;

  case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_f1:
    break;

  case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_f0:
    break;

  case LIBERTY_ATTRENUM_default_connection_class:
    break;

  case LIBERTY_ATTRENUM_default_cell_leakage_power:
    break;

  case LIBERTY_ATTRENUM_date:
    break;

  case LIBERTY_ATTRENUM_current_unit:
    break;

  case LIBERTY_ATTRENUM_comment:
    break;

  case LIBERTY_ATTRENUM_bus_naming_style:
    break;

  default:
    break;
  }
}

static LibGroup*
get_containing_group(LibGroup* group, const char* type)
{
  const struct libGroupMap *lgm;
  LibGroup* gup = group;
  group_enum gt;
  lgm = Perfect_Hash_group::lookup_group_name(type, strlen(type));
  if (lgm)
    gt = lgm->type;
  else {
    sprintf(EB, "BAD type (%s) handed to get_containing_group!", type);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_INTERNAL_SYSTEM_ERROR, EB);
    return NULL;
  }
  while ((gup = gup->owner())) {
    if ((group_enum)gup->typeID() == gt)
      return gup;
  }
  return NULL;
}

static const char *
mystrtod(const char *str, const char **end)
{
  const char *x = str, *y;

  while (*x && (*x == ' ' || *x == '	' || *x == ',' || *x == '\n'
		|| *x == '\\')) {
    x++;
  }
  y = x;

  while (*x && (*x == '+' || *x == '-' || (*x >= '0' && *x <= '9')
		|| *x == 'E' || *x == 'e' || *x == '.')) {
    x++;
  }
  if (y == x) {
    *end = str;
    return str;
  }
  *end = x;
  return x;
}


static int
count_floats_in_string(const char *s)
{
  const char *p, *t;
  int i = 0;

  p = s;

  while (mystrtod(p, &t) != p) {
    i++;
    p = t;
  }
  return i;
}

// static inline int
// count_strings_in_list(LibAttr* attr)
// {
//   LibValIter viter = attr->valueIter();
//   int i = 0;

//   while (viter++) {
//     i++;
//   }
//   return i;
// }

static inline void
printLutRefError (const char* filename, int lineno,
		  const char* gtype, const char* templateName,
		  const char* addMsg)
{
  sprintf(EB, "%s:%d, %s references template %s,",
	  filename, lineno, gtype, templateName);
  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, addMsg);
  errcount++;
}

static void
checkScalarValue (LibAttr* vs, const char* filename, int lineno,
		  const char* gtype, const char* templateName)
{
  LibValIter viter = vs->valueIter();
  LibVal* v = viter++;
  
  if (!v) {
    printLutRefError(filename, lineno, gtype, templateName,
		     "                  so the values attribute should "
		     "have one string specified\n"
		     "                  with a number in it, but does not.");
  } else if (v->type() != LIBERTY__VAL_STRING) {
    printLutRefError(filename, lineno, gtype, templateName,
		     "                  so the values attribute should have one string "
		     "specified\n                  with a number in it, but what is "
		     "there is not a string.");
  } else {
    if (vs->valueCount() != 1 ||
	count_floats_in_string(v->string_val()) != 1) {
      printLutRefError(filename, lineno, gtype, templateName,
		       "                  but the values attribute has more than one "
		       "value or string specified.");
    }
  }
}

static void
checkLutValue(LibAttr* vs, int strcnt, int numPerStr, const char* filename,
	      int lineno, const char* gtype, const char* templateName)
{
  LibValIter viter;
  LibVal* v;
  int numstrs = vs->valueCount();
  char str[500];

  if (numstrs == strcnt) {
    /* now, count the nums in each entry, and make sure they equal numPerStr */
    viter = vs->valueIter();
    for (int i = 0; i < numstrs; i++) {
      v = viter++;
      if (v->type() == LIBERTY__VAL_STRING) {
	int numc = count_floats_in_string(v->string_val());
	if (numc != numPerStr) {
	  sprintf(str, "                  there should be %d string(s) in the values table, each with %d number(s) in it,\n"
		  "                  but in the group's values attribute (ending at line %d) value #%d has %d number(s).",
		  strcnt, numPerStr, vs->lineno(), i+1, numc);
	  printLutRefError(filename, lineno, gtype, templateName, str);
	}
      } else {
	  sprintf(str, "                  there should be %d string(s) in the values table, each with %d number(s) in it,\n"
		  "                  but in the group's values attribute (ending at line %d) value #%d is not a string.",
		  strcnt, numPerStr, vs->lineno(), i+1);
	  printLutRefError(filename, lineno, gtype, templateName, str);
      }
    }
  } else {
    sprintf(str, "                  there should be %d string(s) in the values table, each with %d number(s) in it,\n"
	    "                  but the group's values attribute (ending at line %d) has %d string(s).",
	    strcnt, numPerStr, vs->lineno(), numstrs);
    printLutRefError(filename, lineno, gtype, templateName, str);
  }
}

static void
check_lut_template_arraysize(LibGroup* group, const char* templateType)
{
  LibErrorT err;
  LibNameIter niter;
  const char* str;
  char tmpStr[500];
  LibGroup* rg;
  LibGroup* rg2;

  niter = group->nameIter();
  str = *niter++;
  if (str) {
    const char* gtype = group->type();
    const char* filename = group->filename();
    int lineno = group->lineno();

    if (!strcmp(str, "scalar")) {
      LibAttr* vs;
      /* should be just one value */
      vs = group->findAttr("values", err);

      if (vs) {
	checkScalarValue(vs, filename, lineno, gtype, str);
      } else {
	printLutRefError(filename, lineno, gtype, str,
			 "                  the group does not have a values attribute.");
      }
      return;
    }

    /* find an lu_table_template with this name */
    rg = get_containing_group(group, "library");
    rg2 = rg->findGroup(str, templateType, err);
    if (rg2) {
      LibAttr* ind1;
      LibAttr* ind2;
      LibAttr* ind3;
      LibAttr* vs;

      ind1 = group->findAttr("index_1", err);
      ind2 = group->findAttr("index_2", err);
      ind3 = group->findAttr("index_3", err);
      if (!ind1)
	ind1 = rg2->findAttr("index_1", err);
      if (!ind2)
	ind2 = rg2->findAttr("index_2", err);
      if (!ind3)
	ind3 = rg2->findAttr("index_3", err);

      vs = group->findAttr("values", err);
      if (vs) {
	if (ind1) {
	  int tpl1nums, tpl2nums, tpl3nums;
	  LibValIter viter;
	  LibVal* v;

	  viter = ind1->valueIter();
	  v = viter++;
	  if (v && v->type() == LIBERTY__VAL_STRING) {
	    tpl1nums = count_floats_in_string(v->string_val());

	    if (ind2) {
	      viter = ind2->valueIter();
	      v = viter++;
	      if (v && v->type() == LIBERTY__VAL_STRING) {
		tpl2nums = count_floats_in_string(v->string_val());

		if (ind3) {
		  viter = ind3->valueIter();
		  v = viter++;
		  if (v && v->type() == LIBERTY__VAL_STRING) {
		    tpl3nums = count_floats_in_string(v->string_val());
		    checkLutValue(vs, tpl1nums * tpl2nums, tpl3nums, filename, lineno,
				  gtype, str);
		  } else {
		    sprintf(tmpStr,
			    "                  but the template's index_3 attribute (line %d) does not have a string as a value.",
			    ind3->lineno());
		    printLutRefError(filename, lineno, gtype, str, tmpStr);
		  }
		} else {
		  checkLutValue(vs, tpl1nums, tpl2nums, filename, lineno, gtype, str);
		}
	      } else {
		sprintf(tmpStr,
			"                  but the template's index_2 attribute (line %d) does not have a string as a value.",
			ind2->lineno());
		printLutRefError(filename, lineno, gtype, str, tmpStr);
	      }
	    } else if (ind3) {
		sprintf(tmpStr,
			"                  the template's index_3 was defined (line %d)\n"
			"                  but the template's index_2 attribute cannot be found.",
			ind3->lineno());
		printLutRefError(filename, lineno, gtype, str, tmpStr);
	    } else {
	      checkLutValue(vs, 1, tpl1nums, filename, lineno, gtype, str);
	    }
	  } else {
	    sprintf(tmpStr,
		    "                  but the template's index_1 attribute (line %d) does not have a string as a value.",
		    ind1->lineno());
	    printLutRefError(filename, lineno, gtype, str, tmpStr);
	  }
	} else {
	  if (ind2) {
	    sprintf(tmpStr,
		    "                  the template's index_2 was defined (line %d)\n"
		    "                  but the template's index_1 attribute cannot be found.",
		    ind2->lineno());
	    printLutRefError(filename, lineno, gtype, str, tmpStr);
	  }
	  if (ind3) {
	    sprintf(tmpStr,
		    "                  the template's index_3 was defined (line %d)\n"
		    "                  but the template's index_1 attribute cannot be found.",
		    ind3->lineno());
	    printLutRefError(filename, lineno, gtype, str, tmpStr);
	  }
	}
      } else {
	  printLutRefError(filename, lineno, gtype, str,
			   "                  but the group does not have a values attribute.");
      }
    } else {
      /* it could be a good 'ol polynomial */
      rg2 = rg->findGroup(str, "poly_template", err);
      if (!rg2) {
	  printLutRefError(filename, lineno, gtype, str,
			   "                  but the referenced group could not be found");
      }
    }
  }
}

static void
check_timing_lut_arraysize(LibGroup* group)
{
  check_lut_template_arraysize(group, "lu_table_template");
}

static void
check_power_lut_arraysize(LibGroup* group)
{
  check_lut_template_arraysize(group, "power_lut_template");
}

static void
check_cell_n_prop(LibGroup* group)
{
  LibGroupIter giter = group->groupIter();
  LibGroup* g2;
  group_enum gt;
  int gcell_rise = 0, gcell_fall = 0, gprop_fall = 0, gprop_rise = 0;
  int gcell_rise_ln = 0, gcell_fall_ln = 0, gprop_fall_ln = 0, gprop_rise_ln = 0;
  const char *gcell_rise_fi = NULL, *gcell_fall_fi = NULL,
       *gprop_fall_fi = NULL, *gprop_rise_fi = NULL;
  int gtran_rise = 0, gtran_fall = 0;
  int gtran_rise_ln, gtran_fall_ln;
  const char *gtran_rise_fi, *gtran_fall_fi;

  while ((g2 = giter++)) {
    gt = (group_enum)g2->typeID();
    switch (gt) {
    case LIBERTY_GROUPENUM_cell_rise:
      gcell_rise++;
      gcell_rise_ln = g2->lineno();
      gcell_rise_fi = g2->filename();
      break;

    case LIBERTY_GROUPENUM_cell_fall:
      gcell_fall++;
      gcell_fall_ln = g2->lineno();
      gcell_fall_fi = g2->filename();
      break;
    case LIBERTY_GROUPENUM_rise_propagation:
      gprop_rise++;
      gprop_rise_ln = g2->lineno();
      gprop_rise_fi = g2->filename();
      break;
    case LIBERTY_GROUPENUM_fall_propagation:
      gprop_fall++;
      gprop_fall_ln = g2->lineno();
      gprop_fall_fi = g2->filename();
      break;
    case LIBERTY_GROUPENUM_rise_transition:
      gtran_rise++;
      gtran_rise_ln = g2->lineno();
      gtran_rise_fi = g2->filename();
      break;
    case LIBERTY_GROUPENUM_fall_transition:
      gtran_fall++;
      gtran_fall_ln = g2->lineno();
      gtran_fall_fi = g2->filename();
      break;
    default:
      break;
    }
  }
  if (gcell_rise && gprop_fall) {
    sprintf(EB,
	    "%s:%d and %s:%d: It is illegal to specify both cell_rise and fall_propagation in the same timing group.",
	    gcell_rise_fi, gcell_rise_ln, gprop_fall_fi, gprop_fall_ln);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (gcell_fall && gprop_rise) {
    sprintf(EB,
	    "%s:%d and %s:%d: It is illegal to specify both cell_fall and rise_propagation in the same timing group.",
	    gcell_fall_fi, gcell_fall_ln, gprop_rise_fi, gprop_rise_ln);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (gprop_rise && !gtran_rise) {
    sprintf(EB,
	    "%s:%d: It is illegal to specify rise_propagation without a rise_transition in the same timing group.",
	    gprop_rise_fi, gprop_rise_ln);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (gprop_fall && !gtran_fall) {
    sprintf(EB,
	    "%s:%d: It is illegal to specify fall_propagation without a fall_transition in the same timing group.",
	    gprop_fall_fi, gprop_fall_ln);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (gcell_rise || gcell_fall || gprop_rise || gprop_fall || gtran_rise || gtran_fall) {
    LibAttrIter aiter = group->attrIter();
    LibAttr* attr;
    attr_enum at;
    int gres_rise = 0, gres_fall = 0, gintr_rise = 0, gintr_fall = 0,
      gslope_rise = 0, gslope_fall = 0;
    int gres_rise_ln = 0, gres_fall_ln = 0, gintr_rise_ln = 0,
      gintr_fall_ln = 0, gslope_rise_ln = 0, gslope_fall_ln = 0;
    const char *gres_rise_fi = NULL, *gres_fall_fi = NULL, *gintr_rise_fi = NULL,
      *gintr_fall_fi = NULL, *gslope_rise_fi = NULL, *gslope_fall_fi = NULL;
    
    while ((attr = aiter++)) {
      at = (attr_enum)attr->typeID();
      switch (at) {
      case LIBERTY_ATTRENUM_rise_resistance:
	gres_rise++;
	gres_rise_ln = attr->lineno();
	gres_rise_fi = attr->filename();
	break;
      case LIBERTY_ATTRENUM_fall_resistance:
	gres_fall++;
	gres_fall_ln = attr->lineno();
	gres_fall_fi = attr->filename();
	break;
      case LIBERTY_ATTRENUM_intrinsic_rise:
	gintr_rise++;
	gintr_rise_ln = attr->lineno();
	gintr_rise_fi = attr->filename();
	break;
      case LIBERTY_ATTRENUM_intrinsic_fall:
	gintr_fall++;
	gintr_fall_ln = attr->lineno();
	gintr_fall_fi = attr->filename();
	break;
      case LIBERTY_ATTRENUM_slope_rise:
	gslope_rise++;
	gslope_rise_ln = attr->lineno();
	gslope_rise_fi = attr->filename();
	break;
      case LIBERTY_ATTRENUM_slope_fall:
	gslope_fall++;
	gslope_fall_ln = attr->lineno();
	gslope_fall_fi = attr->filename();
	break;
      default:
	break;
      }
      if (gintr_rise) {
	sprintf(EB,
		"%s:%d: It is illegal to specify intrinsic_rise with a cell_rise(fall),\n"
		"rise(fall)_propagation or rise(fall)_transition in the same timing group.",
		gintr_rise_fi, gintr_rise_ln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
      if (gintr_fall) {
	sprintf(EB,
		"%s:%d: It is illegal to specify intrinsic_fall with a cell_rise(fall),\n"
		"rise(fall)_propagation or rise(fall)_transition in the same timing group.",
		gintr_fall_fi, gintr_fall_ln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
      if (gres_rise) {
	sprintf(EB,
		"%s:%d: It is illegal to specify rise_resistance with a cell_rise(fall),\n"
		"rise(fall)_propagation or rise(fall)_transition in the same timing group.",
		gres_rise_fi, gres_rise_ln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
      if (gres_fall) {
	sprintf(EB,
		"%s:%d: It is illegal to specify fall_resistance with a cell_rise(fall),\n"
		"rise(fall)_propagation or rise(fall)_transition in the same timing group.",
		gres_fall_fi, gres_fall_ln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
      if (gslope_rise) {
	sprintf(EB,
		"%s:%d: It is illegal to specify rise_slope with a cell_rise(fall),\n"
		"rise(fall)_propagation or rise(fall)_transition in the same timing group.",
		gslope_rise_fi, gslope_rise_ln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
      if (gslope_fall) {
	sprintf(EB,
		"%s:%d: It is illegal to specify fall_slope with a cell_rise(fall),\n"
		"rise(fall)_propagation or rise(fall)_transition in the same timing group.",
		gslope_fall_fi, gslope_fall_ln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }

    }
  }
}

static void
check_clear_n_preset(LibGroup* group)
{
  LibErrorT err;

  LibAttr* clear;
  LibAttr* preset;
  LibAttr* var1;
  LibAttr* var2;

  clear = group->findAttr("clear", err);
  preset = group->findAttr("preset", err);

  if (clear && preset) {
    var1 = group->findAttr("clear_preset_var1", err);
    var2 = group->findAttr("clear_preset_var2", err);

    if (!var1 && !var2) {
      sprintf(EB,
	      "%s:%d, Group %s, with both 'clear', and 'preset' attributes,\n"
	      "requires 'clear_preset_var1' and/or 'clear_preset_var2' attributes.",
	      group->filename(), group->lineno(),
	      (str_t)group->type());
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
    }
  }
}

static void
check_members(LibGroup* group)
{
  /* is members specified? Is it first? */
  LibErrorT err;
  LibAttr* mem = group->findAttr("members", err);
  int faln, fgln, mln;

  if (mem) {
    /* a member attr is present */
    LibAttr* attr;
    LibGroup* g2;

    mln = mem->lineno();

    attr = group->firstAttr();
    if (attr) {
      faln = attr->lineno();
    } else {
      faln = -1;
    }
    g2 = group->firstGroup();
    if (g2) {
      fgln = g2->lineno();
    } else {
      fgln = -1;
    }

    if (attr != mem || (fgln > 0 && fgln < mln)) {
      const char* name = group->firstName();
      if (!name)
	name = "<Unnamed>";

      if (fgln > 0 && fgln < mln) {
	sprintf(EB,
		"%s:%d, The %s bundle needs to have its 'members' specified first.\n"
		"The group %s (line %d) is found before 'members'.",
		mem->filename(), mem->lineno(),
		name, (str_t)g2->firstName(), fgln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      }
      if (mem != attr) {
	sprintf(EB,
		"%s:%d, The %s bundle needs to have its 'members' specified first.\n"
		"The attr %s (line %d) is found before 'members'.",
		mem->filename(), mem->lineno(),
		name, (str_t)attr->name(), faln);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      }
      errcount++;
    }
  }
}

#if 0
static void
check_interface_timing(LibGroup* group)
{
  LibErrorT err;
  LibAttr* ift;
  LibGroupIter giter;
  LibGroup* g2;
  group_enum gt;
  const char* name;
  int iftn;
  LibAttr* attr;

  name = group->firstName();

  ift = group->findAttr("interface_timing", err);
  if (ift && ift->type() == LIBERTY__SIMPLE &&
      ift->value()->bool_val() == true) {
    iftn = ift->lineno();

    /* check to make sure the stuff is as it should be... */
    giter = group->groupIter();
    
    while ((g2 = giter++)) {
      gt = (group_enum)g2->typeID();
      switch (gt) {
      case LIBERTY_GROUPENUM_ff:
	sprintf(EB,
		"%s:%d, The cell '%s' has the interface_timing "
		"attribute (line %d), but is not a black box.",
		g2->filename(), g2->lineno(),
		name, iftn);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	sprintf(EB,
		"                  -- The ff group should not be "
		"specified when interface_timing is true.");
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
	break;
	
      case LIBERTY_GROUPENUM_pin:
	attr = g2->findAttr("function", err);
	if (attr) {
	  sprintf(EB,
		  "%s:%d, The cell '%s' has the interface_timing "
		  "attribute (line %d), but is not a black box.",
		  attr->filename(), attr->lineno(),
		  name, iftn);
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  sprintf(EB,
		  "                  -- pin groups should not have the "
		  "function attribute specified when interface_timing is true.");
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  errcount++;
	}
	break;
	
      default:
	break;
      }
    }
  }
}
#endif

static void
check_bus_pin_directions(LibGroup* group)
{				/* group is bus or bundle */
  LibErrorT err;
  LibAttr* bdir;
  LibAttr* pdir;
  group_enum gt;
  const char* bdirstr;
  const char* dirstr;

  bdir = group->findAttr("direction", err);

  if (bdir) {
    LibGroup* g2;
    LibGroupIter giter = group->groupIter();

    bdirstr = bdir->value()->string_val();
    if (!bdirstr || !bdirstr[0])
      return;

    while ((g2 = giter++)) {
      gt = (group_enum)g2->typeID();

      if (gt == LIBERTY_GROUPENUM_pin) {
	pdir = g2->findAttr("direction", err);
	if (pdir) {
	  dirstr = pdir->value()->string_val();
	  if (!dirstr || !dirstr[0])
	    continue;

	  if (strcmp(bdirstr, dirstr)) {
	    sprintf(EB,
		    "%s:%d, The group '%s' has the direction attribute (line %d) '%s', but",
		    pdir->filename(), pdir->lineno(),
		    (str_t)group->firstName(), bdir->lineno(), bdirstr);
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    sprintf(EB,
		    "                  the pin '%s' has the conflicting direction attribute '%s'.",
		    (str_t)g2->firstName(), dirstr);
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    errcount++;
	  }
	}
      }
    }
  }
}

static void
check_pin(LibGroup* group, LibGroup* parent)
{
  LibErrorT err;
  LibAttr* direction = group->findAttr("direction", err);

  if (direction) {
    return;
  }

  if (parent && (group_enum)parent->typeID() == LIBERTY_GROUPENUM_bus
      && parent->findAttr("direction", err)) {
    return;
  }

  sprintf(EB,
	  "%s:%d, pin must have a direction specified",
	  group->filename(),
	  group->lineno());
  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
  errcount++;
}
 
static void
check_driver_types(LibGroup* group)
{				/* group should be a pin, perhaps even bus or bundle  */
  int pullup = 0, pulldown = 0, bushold = 0, opensource = 0, opendrain = 0;
  LibAttr* drivertype;
  LibAttr* direct;
  LibErrorT err;
  const char* str;


  drivertype = group->findAttr("driver_type", err);

  if (!drivertype)	/* if nothing's there, nothing could be wrong! */
    return;

  if (drivertype->type() != LIBERTY__SIMPLE)
    return;

  if (drivertype->value()->type() != LIBERTY__VAL_STRING)
    return;

  str = drivertype->value()->string_val();

  if (!str || !str[0])
    return;

  /* we can do checks now on the drivertype combinations */

  if (strstr(str, "pull_up"))
    pullup = 1;
  if (strstr(str, "pull_down"))
    pulldown = 1;
  if (strstr(str, "open_source"))
    opensource = 1;
  if (strstr(str, "open_drain"))
    opendrain = 1;
  if (strstr(str, "bus_hold"))
    bushold = 1;

  if (pullup && pulldown) {
    sprintf(EB,
	    "%s:%d, The driver_types 'pull_up' and 'pull_down' cannot be specified on the same pin",
	    drivertype->filename(),
	    drivertype->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (pullup && bushold) {
    sprintf(EB,
	    "%s:%d, The driver_types 'pull_up' and 'bus_hold' cannot be specified on the same pin",
	    drivertype->filename(),
	    drivertype->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (pulldown && bushold) {
    sprintf(EB,
	    "%s:%d, The driver_types 'pull_down' and 'bus_hold' cannot be specified on the same pin",
	    drivertype->filename(),
	    drivertype->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (opendrain && opensource) {
    sprintf(EB,
	    "%s:%d, The driver_types 'open_source' and 'open_drain' cannot be specified on the same pin",
	    drivertype->filename(),
	    drivertype->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (opendrain && bushold) {
    sprintf(EB,
	    "%s:%d, The driver_types 'open_drain' and 'bus_hold' cannot be specified on the same pin",
	    drivertype->filename(),
	    drivertype->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (opensource && bushold) {
    sprintf(EB,
	    "%s:%d, The driver_types 'open_source' and 'bus_hold' cannot be specified on the same pin",
	    drivertype->filename(),
	    drivertype->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }

  if (opensource || opendrain) {
    direct = group->findAttr("direction", err);

    if (!direct)	/* if nothing's there, nothing could be wrong! */
      return;

    if (direct->type() != LIBERTY__SIMPLE)
      return;

    if (direct->value()->type() != LIBERTY__VAL_STRING)
      return;

    str = direct->value()->string_val();

    if (!str || !str[0])
      return;

    if (!strcmp(str, "input")) {
      if (opensource) {
	sprintf(EB,
		"%s:%d, The 'open_source' driver_type (line %d) cannot be specified on an input pin",
		direct->filename(),
		direct->lineno(),
		drivertype->lineno());
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      } else {
	sprintf(EB,
		"%s:%d, The 'open_drain' driver_type (line %d) cannot be specified on an input pin",
		direct->filename(),
		direct->lineno(),
		drivertype->lineno());
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
    }
  }
}

static int
get_bus_size(LibGroup* group, const char *busname)
{
  LibErrorT err;
  LibGroup* bus = group->findGroup(busname, "bus", err);

  if (bus) {
    LibAttr* bustype = bus->findAttr("bus_type", err);
    LibGroup* typeb;
    LibGroup* libr;
    LibAttr* bitfrom;
    LibAttr* bitto;
    LibAttr* bitwid;
    int wid = -1;
    if (bustype) {
      const char* str;

      if (bustype->type() != LIBERTY__SIMPLE) {
	return -1;		/* this will already be reported */
      }
      if (bustype->value()->type() != LIBERTY__VAL_STRING) {
	return -1;
      }
      str = bustype->value()->string_val();
      if (!str || !str[0]) {
	return -1;
      }

      /* find the type in the cell or lib */
      typeb = group->findGroup(str, "type", err);
      if (!typeb) {
	libr = get_containing_group(group, "library");
	if (libr) {
	  typeb = libr->findGroup(str, "type", err);
	}
      }
      if (!typeb) {
	return -1;
      }

      bitwid = typeb->findAttr("bit_width", err);
      if (bitwid) {
	LibValType vt;

	if (bitwid->type() != LIBERTY__SIMPLE) {
	  return -1;		/* this will already be reported */
	}
	vt = bitwid->value()->type();
	if (vt == LIBERTY__VAL_STRING) {
	  return -1;
	}
	if (vt == LIBERTY__VAL_INT) {
	  wid = bitwid->value()->int_val();
	} else if (vt == LIBERTY__VAL_DOUBLE) {
	  wid = (int)bitwid->value()->double_val();
	}
      } else {
	bitfrom = typeb->findAttr("bit_from", err);
	bitto = typeb->findAttr("bit_to", err);
	if (bitfrom && bitto) {
	  LibValType vt;
	  int from = 0, to = 0;
	  int big, lit;

	  if (bitfrom->type() != LIBERTY__SIMPLE) {
	    return -1;		/* this will already be reported */
	  }
	  vt = bitfrom->value()->type();
	  if (vt == LIBERTY__VAL_STRING) {
	    return -1;
	  }
	  if (vt == LIBERTY__VAL_INT) {
	    from = bitfrom->value()->int_val();
	  } else if (vt == LIBERTY__VAL_DOUBLE) {
	    from = (int)bitfrom->value()->double_val();
	  }

	  if (bitto->type() != LIBERTY__SIMPLE) {
	    return -1;		/* this will already be reported */
	  }
	  vt = bitto->value()->type();
	  if (vt == LIBERTY__VAL_STRING) {
	    return -1;
	  }
	  if (vt == LIBERTY__VAL_INT) {
	    to = bitto->value()->int_val();
	  } else if (vt == LIBERTY__VAL_DOUBLE) {
	    to = (int)bitto->value()->double_val();
	  }
	  big = (from > to ? from : to);
	  lit = (from > to ? to : from);

	  wid = big - lit + 1;
	} else {
	  return -1;
	}
      }
    }
    return wid;
  }
  return -1;
}

static int
get_bun_size(LibGroup* group, const char *bun)
{
  LibErrorT err;
  LibGroup* bus = group->findGroup(bun, "bundle", err);
  if (bus) {
    LibAttr* bunmem = bus->findAttr("members", err);

    if (bunmem) {
      return bunmem->valueCount();
    }
  }
  return -1;
}


static void
gen_var_list(const char *formula)
{
  /* a very crude boolean algebra formula parser */
  const char *p;
  char *q;
  char varname[100];


  p = formula;
  q = varname;
  while (*p) {
    /* skip over any non-name chars */
    while (*p && (*p == '(' || *p == ')' || *p == ' '
		  || *p == '\t' || *p == '&' || *p == '|' || *p == '*'
		  || *p == '!' || *p == '^' || *p == '+' || *p == '\''))
      p++;

    while (*p
	   && (*p != '(' && *p != ')' && *p != ' ' && *p != '+' && *p != '\''
	       && *p != '\t' && *p != '&' && *p != '|' && *p != '*'
	       && *p != '^' && *p != '!'))
      *q++ = *p++;

    *q = 0;
    if (varname[0]) {
      add_varlist(varname);
    }
    q = varname;
  }
}

static LibGroup*
find_bu_by_function(LibGroup* cellequiv, const char *name)
{
  LibErrorT err;
  LibGroupIter giter = cellequiv->groupIter();
  LibGroup* g2;
  LibAttr* attr;
  LibValType vtype;
  const char* string = 0;
  LibAttrType atype;

  group_enum gt;

  while ((g2 = giter++)) {
    gt = (group_enum)g2->typeID();
    switch (gt) {
    case LIBERTY_GROUPENUM_bus:
    case LIBERTY_GROUPENUM_bundle:
      attr = g2->findAttr("function", err);
      if (attr) {
	atype = attr->type();
	if (atype != LIBERTY__COMPLEX) {
	  vtype = attr->value()->type();
	  switch (vtype) {
	  case LIBERTY__VAL_STRING:
	    string = attr->value()->string_val();
	    if (!strcasecmp(name, string)) {
	      return g2;
	    }
	    break;
	  default:
	    break;
	  }
	}
      }
      break;
    
    default:
      break;
    }
  }
  return NULL;
}

static LibGroup*
find_pin_in_cellequiv(LibGroup* cellequiv, const char *name)
{
  LibGroupIter giter = cellequiv->groupIter();
  LibGroup* g2;
  int found = 0;
  LibNameIter niter;
  const char* str;

  group_enum gt;

  while ((g2 = giter++)) {
    gt = (group_enum)g2->typeID();
    switch (gt) {
    case LIBERTY_GROUPENUM_pin:
      niter = g2->nameIter();
      while (niter) {
	str = *niter++;
	if (!strcmp(name, str)) {
	  found++;
	  break;
	}
      }
      if (found) {
	return g2;
      }
      break;
    
    default:
      break;
    }
  }
  return NULL;
}

static void
check_next_state_formula(LibGroup* cell, const char *formula, int ffbankwidth,
			 LibAttr* attr)
{
  LibErrorT err;
  var_list *v;
  LibGroup* bus;

  destroy_varlist();
  gen_var_list(formula);

  for (v = master_var_list; v; v = v->next) {
    /* assume: every variable is either a bus or bundle;
       a variable may match the function of a bus or bundle;
       the width of the bus must match the width of the ff_bank */
    /* first, try to match the bus or bun */
    bus = cell->findGroup(v->name, "bus", err);
    if (!bus) {
      bus = cell->findGroup(v->name, "bundle", err);
      if (bus) {
	int wid;

	wid = get_bun_size(cell, v->name);
	if (wid == -1) {
	  sprintf(EB,
		  "%s:%d, The 'members' attribute was not found in the bundle. Cannot make bus width correspondence checks.",
		  bus->filename(), bus->lineno());
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  errcount++;
	  continue;
	} else {
	  if (wid != ffbankwidth) {
	    sprintf(EB,
		    "%s:%d, This ff_bank, with width %d, does not match the referenced bundle %s (line %d), whose width is %d",
		    attr->filename(),
		    attr->lineno(), ffbankwidth, v->name,
		    bus->lineno(), wid);
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    errcount++;
	    continue;
	  }
	}
	continue;
      }
    }

    if (bus) {
      int wid = get_bus_size(cell, v->name);

      if (wid != ffbankwidth) {
	sprintf(EB,
		"%s:%d, This ff_bank, with width %d, does not match the referenced bus %s (line %d), whose width is %d",
		attr->filename(), attr->lineno(),
		ffbankwidth, v->name, bus->lineno(), wid);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
	continue;
      }
    } else {
      group_enum gt;
      const char* str;
      int wid;
      LibGroup* pinn;

      bus = find_bu_by_function(cell, v->name);

      if (!bus) {
	pinn = find_pin_in_cellequiv(cell, v->name);

	if (!pinn) {
	  sprintf(EB,
		  "%s:%d, This ff_bank, with width %d, references the object %s, which I cannot find.",
		  attr->filename(),
		  attr->lineno(), ffbankwidth, v->name);
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  errcount++;
	  continue;
	} else
	  continue;		/* not an error */
      }

      str = bus->firstName();
      if (str && str[0]) {
	gt = (group_enum)bus->typeID();
	switch (gt) {
	case LIBERTY_GROUPENUM_bus:
	  wid = get_bus_size(cell, str);
	  break;

	case LIBERTY_GROUPENUM_bundle:
	  wid = get_bun_size(cell, str);
	  break;

	default:
	  wid = 0;
	  break;
	}
	if (wid != ffbankwidth) {
	  sprintf(EB,
		  "%s:%d, This ff_bank, with width %d, does not match the referenced bus %s with function %s (line %d), whose width is %d",
		  attr->filename(),
		  attr->lineno(), ffbankwidth, str, v->name,
		  bus->lineno(), wid);
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  errcount++;
	  continue;
	}
      }
    }
  }
  destroy_varlist();
}

static void
check_ff_bank_widths(LibGroup* group)
{				/* called for ff_banks */
  LibErrorT err;
  const char* str;
  LibNameIter niter;
  int ffbankwidth = 0;
  LibAttr* attr;
  LibGroup* cell;

  niter = group->nameIter();
  str = niter ? *niter++ : NULL;
  if (str) {
    str = niter ? *niter++ : NULL;
    if (str) {
      str = niter ? *niter++ : NULL;
      if (str) {
	ffbankwidth = atoi(str);
      }
    }
  }
  if (ffbankwidth == 0) {
    sprintf(EB,
	    "%s:%d, Could not retrieve ff_bank width from file, or it has value 0. Cannot make bus width correspondence checks.",
	    group->filename(), group->lineno());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    return;
  }
  attr = group->findAttr("next_state", err);
  if (attr) {
    if (attr->type() != LIBERTY__SIMPLE) {
      return;			/* this will already be reported */
    }
    if (attr->value()->type() != LIBERTY__VAL_STRING) {
      sprintf(EB,
	      "%s:%d, The 'next_state' attribute is not of string type. Cannot make bus width correspondence checks.",
	      attr->filename(), attr->lineno());
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      return;
    }
    str = attr->value()->string_val();
    if (!str || !str[0]) {
      sprintf(EB,
	      "%s:%d, The 'next_state' attribute is an empty string. Cannot make bus width correspondence checks.",
	      attr->filename(), attr->lineno());
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      return;
    }

    /* go find this referenced bus */
    cell = get_containing_group(group, "test_cell");
    if (!cell)
      cell = get_containing_group(group, "model");
    if (!cell)
      cell = get_containing_group(group, "scaled_cell");
    if (!cell)
      cell = get_containing_group(group, "cell");
    if (cell) {
      check_next_state_formula(cell, str, ffbankwidth, attr);
    }
  }
  return;
}

static void
check_bus_type(LibGroup* bus)
{
  LibErrorT err;
  const char* str;
  LibGroup* cell;
  LibAttr* bustype = bus->findAttr("bus_type", err);
  LibGroup* typeb;
  LibGroup* libr;

  if (bustype) {
    if (bustype->type() != LIBERTY__SIMPLE) {
      return;			/* this will already be reported */
    }
    if (bustype->value()->type() != LIBERTY__VAL_STRING) {
      return;
    }
    str = bustype->value()->string_val();
    if (!str || !str[0]) {
      sprintf(EB, "%s:%d, The 'bus_type' attribute is an empty string.",
	      bustype->filename(),
	      bustype->lineno());
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      return;
    }

    /* find the type in the cell or lib */
    cell = get_containing_group(bus, "test_cell");
    if (!cell)
      cell = get_containing_group(bus, "model");
    if (!cell)
      cell = get_containing_group(bus, "scaled_cell");
    if (!cell)
      cell = get_containing_group(bus, "cell");

    typeb = cell->findGroup(str, "type", err);
    if (!typeb) {
      libr = get_containing_group(bus, "library");
      if (libr) {
	typeb = libr->findGroup(str, "type", err);
      }
    }
    if (!typeb) {
      sprintf(EB,
	      "%s:%d, The 'bus_type' references the type '%s', which cannot be found in the containing cell or library.",
	      bustype->filename(),
	      bustype->lineno(), str);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      return;
    }
  } else {
    sprintf(EB, "%s:%d, The 'bus_type' attribute is missing in bus '%s'.",
	    bus->filename(),
	    bus->lineno(), (str_t)bus->firstName());
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    return;
  }

}

static void
find_all_refs_to_template(LibGroup* group, const char *name, int num)
{
  LibGroupIter giter = group->groupIter();
  LibGroup* g2;
  LibNameIter niter;
  const char* str;

  while ((g2 = giter++)) {
    if (!strcmp(g2->type(), "lu_table_template"))
      continue;

    niter = g2->nameIter();
    while (niter) {
      str = *niter++;
      if (!strcmp(name, str)) {
	if (!num) {
	  sprintf(EB,
		  "         Here is a list of references to this invalid template:");
	  (*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_SYNTAX_ERROR, EB);
	}

	num++;
	sprintf(EB, "           %s:%d, group '%s'", g2->filename(),
		g2->lineno(), (str_t)g2->type());
	(*MsgPrinter) (LIB_SEVERITY_NOTE, LIB_SYNTAX_ERROR, EB);
	break;
      }
    }
    find_all_refs_to_template(g2, name, num);	/* recursively descend */
  }
}

static void
check_index_x(const char* indnam, LibGroup* group)
{
  LibErrorT err;
  LibGroup* lib;
  LibAttr* index_x = group->findAttr(indnam, err);
  LibVal* val;
  const char* string, *p;
  char *t;
  double x;

  if (index_x) {
    int inval;

    /* should be a string with a list of comma/space sep'd numbers */
    if (index_x->type() != LIBERTY__COMPLEX) {
      return;			/* this will already be reported */
    }

    val = index_x->firstValue();
    if (!val) {
      sprintf(EB, "%s:%d, The '%s' attribute has no values.",
	      index_x->filename(),
	      index_x->lineno(), indnam);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      return;
    }

    if (val->type() != LIBERTY__VAL_STRING) {
      sprintf(EB, "%s:%d, The '%s' attribute should be a string.",
	      index_x->filename(),
	      index_x->lineno(), indnam);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      return;
    }
    string = val->string_val();
    if (!string || !string[0]) {
      sprintf(EB, "%s:%d, The '%s' attribute is an empty string.",
	      index_x->filename(),
	      index_x->lineno(), indnam);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
      return;
    }

    p = string;
    inval = 0;
    while ((x = strtod(p, &t)), t != p) {
      if (x < 0.0) {
	sprintf(EB,
		"%s:%d, The '%s' attribute contains the negative value '%g'. The template '%s' is therefore invalid.",
		index_x->filename(),
		index_x->lineno(), indnam, x,
		(str_t)group->firstName());
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
	inval = 1;
      }
      if (t && *t == 0)
	break;
      p = t + 1;
    }

    if (inval) {
      lib = get_containing_group(group, "library");
      find_all_refs_to_template(lib, group->firstName(), 0);
    }
  }
}

static void
check_lut_template(LibGroup* group)
{
  check_index_x("index_1", group);
  check_index_x("index_2", group);
}


static void
group_specific_checks(LibGroup* group, LibGroup* parent)
{
  LibErrorT err;
  group_enum grouptype = (group_enum)group->typeID();
  LibAttr* attr;

  switch (grouptype) {
  case LIBERTY_GROUPENUM_lut:
    break;

  case LIBERTY_GROUPENUM_wire_load_table:
    break;

  case LIBERTY_GROUPENUM_wire_load_selection:
    break;

  case LIBERTY_GROUPENUM_interconnect_delay:
    break;

  case LIBERTY_GROUPENUM_wire_load:
    break;

  case LIBERTY_GROUPENUM_timing_range:
    break;

  case LIBERTY_GROUPENUM_scaling_factors:
    break;

  case LIBERTY_GROUPENUM_scaled_cell:
    break;

  case LIBERTY_GROUPENUM_rise_transition_degradation:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_rise_net_delay:
    break;

  case LIBERTY_GROUPENUM_power_supply:
    break;

  case LIBERTY_GROUPENUM_power_lut_template:
    break;

  case LIBERTY_GROUPENUM_output_voltage:
    break;

  case LIBERTY_GROUPENUM_operating_conditions:
    break;

  case LIBERTY_GROUPENUM_model:
    break;

  case LIBERTY_GROUPENUM_lu_table_template:
    check_lut_template(group);
    break;

  case LIBERTY_GROUPENUM_input_voltage:
    break;

  case LIBERTY_GROUPENUM_fall_transition_degradation:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_fall_net_delay:
    break;

  case LIBERTY_GROUPENUM_em_lut_template:
    break;

  case LIBERTY_GROUPENUM_type:
    break;

  case LIBERTY_GROUPENUM_test_cell:
    break;

  case LIBERTY_GROUPENUM_statetable:
    break;

  case LIBERTY_GROUPENUM_state:
    break;

  case LIBERTY_GROUPENUM_seq_bank:
    check_clear_n_preset(group);
    break;

  case LIBERTY_GROUPENUM_seq:
    check_clear_n_preset(group);
    break;

  case LIBERTY_GROUPENUM_routing_track:
    break;

  case LIBERTY_GROUPENUM_mode_value:
    break;

  case LIBERTY_GROUPENUM_mode_definition:
    break;

  case LIBERTY_GROUPENUM_memory:
    break;

  case LIBERTY_GROUPENUM_leakage_power:
    break;

  case LIBERTY_GROUPENUM_latch_bank:
    check_clear_n_preset(group);
    break;

  case LIBERTY_GROUPENUM_latch:
    check_clear_n_preset(group);
    break;

  case LIBERTY_GROUPENUM_generated_clock:
    break;

  case LIBERTY_GROUPENUM_ff_bank:
    check_clear_n_preset(group);
    check_ff_bank_widths(group);
    break;

  case LIBERTY_GROUPENUM_ff:
    check_clear_n_preset(group);
    break;

  case LIBERTY_GROUPENUM_bus:
    check_bus_pin_directions(group);
    check_bus_type(group);
    break;

  case LIBERTY_GROUPENUM_tlatch:
    break;

  case LIBERTY_GROUPENUM_rise_transition:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_rise_propagation:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_rise_constraint:
    break;

  case LIBERTY_GROUPENUM_retaining_rise:
    break;

  case LIBERTY_GROUPENUM_retaining_fall:
    break;

  case LIBERTY_GROUPENUM_fall_transition:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_fall_propagation:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_fall_constraint:
    break;

  case LIBERTY_GROUPENUM_cell_rise:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_cell_fall:
    check_timing_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_cell_degradation:
    break;

  case LIBERTY_GROUPENUM_timing:
    check_cell_n_prop(group);
    break;

  case LIBERTY_GROUPENUM_pin:
    check_pin(group, parent);
    check_driver_types(group);
    break;

  case LIBERTY_GROUPENUM_minimum_period:
    break;

  case LIBERTY_GROUPENUM_min_pulse_width:
    break;

  case LIBERTY_GROUPENUM_memory_write:
    break;

  case LIBERTY_GROUPENUM_memory_read:
    break;

  case LIBERTY_GROUPENUM_rise_power:
    check_power_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_power:
    check_power_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_fall_power:
    check_power_lut_arraysize(group);
    break;

  case LIBERTY_GROUPENUM_internal_power:

    /* if there is a value here, then check it; otherwise the power/rise_power/fall_power 
       groups will be there */
    attr = group->findAttr("values", err);
    if (attr) {
      check_power_lut_arraysize(group);
    }
    break;

  case LIBERTY_GROUPENUM_em_max_toggle_rate:
    break;

  case LIBERTY_GROUPENUM_electromigration:
    break;

  case LIBERTY_GROUPENUM_bundle:
    check_members(group);
    check_bus_pin_directions(group);
    break;

  case LIBERTY_GROUPENUM_cell:
    //    check_interface_timing(group);
    break;

  case LIBERTY_GROUPENUM_library:
    break;

  default:
    break;

  }
}

static inline void
print_name_error(const char* numNameConsStr, int numName, const char* gtype,
		 const char* gf, int gl)
{
  sprintf(EB,
	  "%s:%d, The %s group in this context should have %s name(s), "
	  "but %d name(s) is(are) there.",
	  gf, gl, gtype, numNameConsStr, numName);
  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
  errcount++;
}
		 
static void
check_name_constraints(libsynt_name_constraints nmCons, int nc,
		       const char* gtype, const char* gf, int gl)
{
 
  switch (nmCons) {
  case SYNTAX_GNAME_NONE:
    if (nc > 0) {
      print_name_error("0", nc, gtype, gf, gl);
    }
    return;
  case SYNTAX_GNAME_ONE:
    if (nc != 1) {
      print_name_error("1", nc, gtype, gf, gl);
    }
    return;
  case SYNTAX_GNAME_TWO:
    if (nc != 2) {
      print_name_error("2", nc, gtype, gf, gl);
    }
    return;
  case SYNTAX_GNAME_THREE:
    if (nc != 3) {
      print_name_error("3", nc, gtype, gf, gl);
    }
    return;
  case SYNTAX_GNAME_NONE_OR_ONE:
    if (nc != 1 && nc != 0) {
      print_name_error("0 or 1", nc, gtype, gf, gl);
    }
    return;
  case SYNTAX_GNAME_ONE_OR_MORE:
    if (nc == 0) {
      print_name_error("1 or more", nc, gtype, gf, gl);
    }
    return;
  }
}
	       
static void
check_group_correspondence(LibGroup* g, LibGroup* parent, libsynt_group_info * gi)
{
  int nc;
  LibNameIter niter;
  LibAttr* attr;
  LibAttrIter aiter;
  LibGroup* group;
  LibGroupIter giter;

  int gl = g->lineno();
  const char *gf = g->filename();
  char gtype[100];

  /* check all group attributes, including names, etc. to constraints */

  if (gi && gi->ref)
    gi = gi->ref;

  strcpy(gtype, g->type());

  /* names */
  nc = 0;
  niter = g->nameIter();
  while (niter++) {
    nc++;
  }
  if (gi) {
    check_name_constraints (gi->name_constraints, nc, gtype, gf, gl);
  }

  /* attrs */
  aiter = g->attrIter();
  while ((attr = aiter++)) {
    const char* anam = attr->name();
    si2drObjectIdT toid;
    int al = attr->lineno();
    const char *af = attr->filename();

    toid.v1 = NULL;
    if (gi)
      liberty_hash_lookup(gi->attr_hash, anam, &toid);
    if (toid.v1 == NULL) {
      /* defines could explain the presence of a simple attr that
         is not found in the group attrs */
      LibDefine* def = LibParse::findDefine(anam);

      if (!def) {
	sprintf(EB,
		"%s:%d, The %s attribute is not allowed in the %s group in this context.",
		af, al, anam, gtype);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      } else {
	/* check to make sure the define's constraints are being obeyed */
	const char* allowed_group = def->groupType();
	LibAttrType t = attr->type();
	char buf[1000], *p, *q;
	strcpy(buf, allowed_group);
	q = buf;
	int found = 0;
	strcat(buf, "|");
	if (strcmp(gtype, allowed_group)) {
	  while (*q && (p = strchr(q, '|'))) {
	    *p = 0;
	    if (!strcmp(q, gtype)) {
	      found = 1;
	      break;
	    }
	    q = p + 1;
	  }
	  if (!found) {
	    sprintf(EB,
		    "%s:%d, The %s attribute was defined to be used in the %s group, not the %s group.",
		    af, al, anam, allowed_group, gtype);
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    errcount++;
	  }
	}
	if (t != LIBERTY__SIMPLE) {
	  sprintf(EB,
		  "%s:%d, The %s attribute was defined and needs to used as a simple attribute.",
		  af, al, anam);
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  errcount++;
	}
#ifdef DONT_VERIFY_DEFINE_ATTRIBUTE_TYPE
	else {
	  LibValType vt;
	  vt = attr->value()->type();
	  if (valtype == LIBERTY__VAL_STRING && vt != LIBERTY__VAL_STRING) {
	    sprintf(EB, "%s:%d, The %s attribute should be of type String.",
		    af, al, anam);
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    errcount++;
	  }
	  if (valtype != LIBERTY__VAL_STRING && vt == LIBERTY__VAL_STRING) {
	    sprintf(EB,
		    "%s:%d, The %s attribute should not be of type String.",
		    af, al, anam);
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    errcount++;
	  }
	}
#endif
      }
    } else {
      check_attr_correspondence(attr, (libsynt_attribute_info *) toid.v1);
    }
  }

  /* subgroups */
  giter = g->groupIter();
  while ((group = giter++)) {
    const char* gt = group->type();
    si2drObjectIdT toid;
    int al = group->lineno();
    const char *af = group->filename();

    toid.v1 = NULL;

    if (gi)
      liberty_hash_lookup(gi->group_hash, gt, &toid);

    if (toid.v1 == NULL) {
      /* check to see if a define_group is defined for this situation */
      LibDefine* def = LibParse::findDefine(gt);
      if (!def) {
	sprintf(EB,
		"%s:%d, The %s group is not allowed in the %s group in this context.",
		af, al, gt, gtype);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      } else {
	const char* allowed_group = def->groupType();
	char buf[1000], *p, *q;
	strcpy(buf, allowed_group);
	q = buf;
	int found = 0;
	strcat(buf, "|");
	if (strcmp(gtype, allowed_group)) {
	  while (*q && (p = strchr(q, '|'))) {
	    *p = 0;
	    if (!strcmp(q, gtype)) {
	      found = 1;
	      break;
	    }
	    q = p + 1;
	  }
	  if (!found) {
	    sprintf(EB,
		    "%s:%d, The %s group is allowed in the %s group, but not allowed in the %s group in this case.",
		    af, al, gt, allowed_group, gtype);
	    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	    errcount++;
	  } else {
	    check_group_correspondence(group, g, (libsynt_group_info *) toid.v1);
	  }
	} else {
	  check_group_correspondence(group, g, (libsynt_group_info *) toid.v1);
	}
      }
    } else {
      check_group_correspondence(group, g, (libsynt_group_info *) toid.v1);
    }
  }
  if (gi) {
    group_specific_checks(g, parent);
  }
}

static void
check_value_range(libsynt_attribute_info* ai, LibVal* v, const char* aname,
		  const char* af, int al)
{
  double min = DBL_MIN;
  double max = DBL_MAX;
  bool   hasRange = false;

  switch (ai->type) {
  case SYNTAX_ATTRTYPE_INT:
    if (ai->u.intcon) {
      if (ai->u.intcon->greater_specd) {
	hasRange = true;
	min = ai->u.intcon->greater_than;
      }
      if (ai->u.intcon->less_specd) {
	hasRange = true;
        max = ai->u.intcon->less_than;
      }
    }
    break;
  case SYNTAX_ATTRTYPE_FLOAT:
    if (ai->u.floatcon) {
      if (ai->u.floatcon->greater_specd) {
	hasRange = true;
	min = ai->u.floatcon->greater_than;
      }
      if (ai->u.floatcon->less_specd) {
	hasRange = true;
	max = ai->u.floatcon->less_than;
      }
    }
    break;
  case SYNTAX_ATTRTYPE_ENUM:
    {
      bool found = 0;
      const char *string = v->string_val();
      for (int i = 0; i < ai->u.stringenum->size; i++) {
	if (!strcmp(ai->u.stringenum->array[i], string)) {
	  found = 1;
	  break;
	}
      }
      if (!found) {
	sprintf(EB,
		"%s:%d, Attribute %s has value %s, which does not match any of the following values:\n     ",
		af, al, aname, string);

	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	for (int i = 0; i < ai->u.stringenum->size; i++) {
	  printf("%s  ", ai->u.stringenum->array[i]);
	}
	printf ("\n");
	errcount++;
      }
      return;
    }

  default:
    return;
  }
  if (!hasRange) {
    return;
  }
  double val = v->double_val();
  if (val < min) {
    sprintf(EB, "%s:%d, Attribute %s has value %g, but it should be greater than %g.",
	    af, al, aname, val, min);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
  if (val > max) {
    sprintf(EB, "%s:%d, Attribute %s has value %g, but it should be less than %g.",
	    af, al, aname, val, max);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
  }
}

static void
check_int_attr(libsynt_attr_type t, const char* aname,
	       const char* af, int al)
{
  const char* typestr = NULL;

  switch (t) {
  case SYNTAX_ATTRTYPE_INT:
    break;

  case SYNTAX_ATTRTYPE_FLOAT:
    break;

  case SYNTAX_ATTRTYPE_VIRTUAL:
    break;
  case SYNTAX_ATTRTYPE_STRING:
    typestr = "a string";
  case SYNTAX_ATTRTYPE_ENUM:
    if (!typestr) typestr = "an enumerated string"; 
  case SYNTAX_ATTRTYPE_BOOLEAN:
    if (!typestr) typestr = "a boolean";
    sprintf(EB,
	    "%s:%d, Attribute %s should be %s Attribute, but is integer instead.",
	    af, al, aname, typestr);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    break;

  default:
    break;
	
  }
}

static void
check_string_attr (libsynt_attr_type t, const char* string,
		   const char* aname, const char* af, int al)
{
  char* x;

  switch (t) {
  case SYNTAX_ATTRTYPE_STRING:
  case SYNTAX_ATTRTYPE_VIRTUAL:
  case SYNTAX_ATTRTYPE_ENUM:
    break;

  case SYNTAX_ATTRTYPE_BOOLEAN:
    if (strcasecmp(string, "false") && strcasecmp(string, "true")) {
      sprintf(EB,
	      "%s:%d, Attribute %s should be a boolean Attribute, but is a string (%s) instead.",
	      af, al, aname, string);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
    }
    break;

  case SYNTAX_ATTRTYPE_INT:
  case SYNTAX_ATTRTYPE_FLOAT:
    strtod(string, &x);
    if (*x != 0) {
      if (!isa_formula(string)) {
	sprintf(EB,
		"%s:%d, Attribute %s should be an integer or float Attribute, but is a string (%s) instead.",
		af, al, aname, string);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
    }
    break;
    
  default:
    break;

  }
}

static void
check_double_attr(libsynt_attr_type t,
		  const char* aname, const char* af, int al)
{
  const char* typestr = NULL;

  switch (t) {
  case SYNTAX_ATTRTYPE_VIRTUAL:
    break;

  case SYNTAX_ATTRTYPE_ENUM:
    typestr = "an enumerated string";
  case SYNTAX_ATTRTYPE_STRING:
    if (!typestr) typestr = "a string";
  case SYNTAX_ATTRTYPE_BOOLEAN:
    if (!typestr) typestr = "a boolean";
    sprintf(EB,
	    "%s:%d, Attribute %s should be %s Attribute, but is float instead.",
	    af, al, aname, typestr);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    break;

  case SYNTAX_ATTRTYPE_INT:
    break;
    
  case SYNTAX_ATTRTYPE_FLOAT:
    break;

  default:
    break;
	
  }
}

static void
check_bool_attr(libsynt_attr_type t, const char* aname,
		const char* af, int al)
{
  const char* typestr = NULL;

  switch (t) {
  case SYNTAX_ATTRTYPE_BOOLEAN:
  case SYNTAX_ATTRTYPE_VIRTUAL:
    break;

  case SYNTAX_ATTRTYPE_STRING:
    typestr = "a string";
  case SYNTAX_ATTRTYPE_INT:
  case SYNTAX_ATTRTYPE_FLOAT:
    if (!typestr) typestr = "an integer/float";
  case SYNTAX_ATTRTYPE_ENUM:
    if (!typestr) typestr = "an enumerated string";
    sprintf(EB,
	    "%s:%d, Attribute %s should be %s Attribute, but is boolean instead.",
	    af, al, aname, typestr);
    (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
    errcount++;
    break;

  default:
    break;
    
  }
}

static void
check_value_type(libsynt_attr_type t, LibVal* v, const char* aname,
		 const char* af, int al)
{
  const char* string;

  LibValType vt = v->type();
  switch (vt) {
  case LIBERTY__VAL_INT:
    check_int_attr(t, aname, af, al);
    break;
  case LIBERTY__VAL_DOUBLE:
    check_double_attr(t, aname, af, al);
    break;
  case LIBERTY__VAL_STRING:
    string = v->string_val();
    check_string_attr(t, string, aname, af, al);
    break;
  case LIBERTY__VAL_BOOLEAN:
    check_bool_attr(t, aname, af, al);
    break;
  default:
    break;
  }
}

static void
check_attr_correspondence(LibAttr* attr, libsynt_attribute_info * ai)
{
  LibAttrType t = attr->type();
  int al = attr->lineno();
  const char* af = attr->filename();
  const char* aname = attr->name();
  LibVal* v;

  if (t == LIBERTY__SIMPLE) {
    if (ai->type == SYNTAX_ATTRTYPE_COMPLEX ||
	ai->type == SYNTAX_ATTRTYPE_COMPLEX_UNKNOWN ||
	ai->type == SYNTAX_ATTRTYPE_COMPLEX_LIST) {
      sprintf(EB,
	      "%s:%d, Attribute %s should be complex, but is simple here.",
	      af, al, aname);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
    } else {
      check_value_type(ai->type, attr->value(), aname, af, al);
      check_value_range(ai, attr->value(), aname, af, al);
    }
  } else if (t == LIBERTY__COMPLEX) {
    if (ai->type != SYNTAX_ATTRTYPE_COMPLEX &&
	ai->type != SYNTAX_ATTRTYPE_COMPLEX_UNKNOWN &&
	ai->type != SYNTAX_ATTRTYPE_COMPLEX_LIST) {
      sprintf(EB,
	      "%s:%d, Attribute %s should be simple, but is complex here.",
	      af, al, aname);
      (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
      errcount++;
    } else {
      LibValIter viter = attr->valueIter();
      libsynt_argument *arg = ai->u.args;
      int argnum;

      argnum = 1;
      while ((v = viter++)) {
	if (!arg) {
	  sprintf(EB, "%s:%d, %s has too many arguments.", af, al, aname);
	  (*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	  errcount++;
	}
	if (arg->type == SYNTAX_ATTRTYPE_COMPLEX_UNKNOWN
	    || arg->type == SYNTAX_ATTRTYPE_COMPLEX_LIST) {
	  arg = arg->next;
	  break;		/* that's all folks */
	}
	check_value_type(arg->type, v, aname, af, al);
	argnum++;
	if (arg)
	  arg = arg->next;
      }
      if (arg) {
	sprintf(EB, "%s:%d, %s has too few arguments.", af, al, aname);
	(*MsgPrinter) (LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, EB);
	errcount++;
      }
    }
  }
  attribute_specific_checks(attr);
}
